@@ -14,11 +14,12 @@ use derive_builder::{Builder, UninitializedFieldError};
 use getset::Getters;
 
 use crate::{
-    self as neli, FromBytes, FromBytesWithInput, FromBytesWithInputBorrowed, Header, Size, ToBytes,
+    self as neli,
     attr::{AttrHandle, Attribute},
     consts::rtnl::*,
     err::{DeError, SerError},
     types::{Buffer, RtBuffer},
+    FromBytes, FromBytesWithInput, FromBytesWithInputBorrowed, Header, Size, ToBytes,
 };
 
 /// Struct representing interface information messages
@@ -89,7 +90,7 @@ pub struct Ifaddrmsg {
     /// Interface address scope
     #[getset(get = "pub")]
     ifa_scope: RtScope,
-    /// Interface address index
+    /// Index of the interface that the address is associated with
     #[getset(get = "pub")]
     ifa_index: libc::c_uint,
     /// Payload of [`Rtattr`]s
@@ -236,6 +237,30 @@ pub struct Tcmsg {
     rtattrs: RtBuffer<Tca, Buffer>,
 }
 
+/// Struct representing VLAN Flags
+#[derive(Builder, Getters, Debug, Size, ToBytes, FromBytes)]
+#[builder(pattern = "owned")]
+pub struct IflaVlanFlags {
+    /// Flags
+    #[getset(get = "pub")]
+    flags: VlanFlags,
+    /// Mask
+    #[getset(get = "pub")]
+    mask: VlanFlags,
+}
+
+/// Struct representing VLAN QoS mapping
+#[derive(Builder, Getters, Debug, Size, ToBytes, FromBytes)]
+#[builder(pattern = "owned")]
+pub struct IflaVlanQosMapping {
+    /// From
+    #[getset(get = "pub")]
+    from: u32,
+    /// To
+    #[getset(get = "pub")]
+    to: u32,
+}
+
 /// Struct representing route netlink attributes
 #[derive(Builder, Getters, Clone, Debug, Size, ToBytes, FromBytes, Header)]
 #[neli(header_bound = "T: RtaType")]
@@ -322,7 +347,7 @@ where
 
     /// Return an [`AttrHandle`] for
     /// attributes nested in the given attribute payload.
-    pub fn get_attr_handle<R>(&self) -> Result<RtAttrHandle<R>, DeError>
+    pub fn get_attr_handle<R>(&self) -> Result<RtAttrHandle<'_, R>, DeError>
     where
         R: RtaType,
     {
@@ -367,7 +392,7 @@ where
 {
     /// Get the payload of an attribute as a handle for parsing
     /// nested attributes.
-    pub fn get_nested_attributes<S>(&self, subattr: T) -> Result<RtAttrHandle<S>, DeError>
+    pub fn get_nested_attributes<S>(&self, subattr: T) -> Result<RtAttrHandle<'_, S>, DeError>
     where
         S: RtaType,
     {
